#define BUILD_NUMBER \
	"build-24915695"
#define BUILD_NUMBER_NUMERIC \
	24915695
#define BUILD_NUMBER_NUMERIC_STRING \
	"24915695"
#define PRODUCT_BUILD_NUMBER \
	"product-build-52705"
#define PRODUCT_BUILD_NUMBER_NUMERIC \
	52705
#define PRODUCT_BUILD_NUMBER_NUMERIC_STRING \
	"52705"
