/*********************************************************
 * Copyright (c) 2019-2020 VMware, Inc. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation version 2.1 and no later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the Lesser GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.
 *
 *********************************************************/

#ifndef _GUESTSTORECLIENTLIB_VERSION_H_
#define _GUESTSTORECLIENTLIB_VERSION_H_

#include "vm_tools_version.h"
#define GUESTSTORECLIENTLIB_VERSION_COMMAS   TOOLS_VERSION_EXT_CURRENT_CSV
#define GUESTSTORECLIENTLIB_VERSION_STRING   TOOLS_VERSION_EXT_CURRENT_STR

#endif /* _GUESTSTORECLIENTLIB_VERSION_H_ */
